from enum import Enum, auto
from typing import Union, overload, Literal, Tuple


class EstimateTimeTopic(Enum):
    unknown = auto()  # None
    duration = auto()  # int, 单位：分钟
    estimate = auto()  # (int, int) 单位：分钟


estimate_time_topic_literal_t = Literal[EstimateTimeTopic.unknown, EstimateTimeTopic.duration, EstimateTimeTopic.estimate]

estimate_time_msg_t = Union[
    Tuple[Literal[EstimateTimeTopic.unknown], None],
    Tuple[Literal[EstimateTimeTopic.duration], int],
    Tuple[Literal[EstimateTimeTopic.estimate], Tuple[int, int]],
]


class EstimateTime:
    message: estimate_time_msg_t

    @overload
    def __init__(self, topic: Literal[EstimateTimeTopic.unknown]) -> None: ...
    @overload
    def __init__(self, topic: Literal[EstimateTimeTopic.duration], payload: int) -> None: ...
    @overload
    def __init__(self, topic: Literal[EstimateTimeTopic.estimate], payload: Tuple[int, int]) -> None: ...

    def __init__(self, topic, payload=None):
        self.message = topic, payload

    def topic(self) -> estimate_time_topic_literal_t:
        return self.message[0]

    @overload
    def payload(self, topic: Literal[EstimateTimeTopic.unknown]) -> None: ...
    @overload
    def payload(self, topic: Literal[EstimateTimeTopic.duration]) -> int: ...
    @overload
    def payload(self, topic: Literal[EstimateTimeTopic.estimate]) -> Tuple[int, int]: ...

    def payload(self, topic):
        return self.message[1]



from typing import List, Iterator, NamedTuple
from datetime import date as Date
import re


def clear_comment(lines: List[str]) -> Iterator[str]:
    for line in lines:
        if '#' in line:
            yield line.split('#', 1)[0]
        else:
            yield line


class block_tuple_t(NamedTuple):
    start_line_no: int
    end_line_no: int


def split_into_blocks(lines: List[str]) -> Iterator[block_tuple_t]:
    """
    按空行将lines分隔成若干个block
    """
    start_line_no = -1
    end_line_no = -1
    for cur_line_no in range(0, len(lines)):
        line = lines[cur_line_no]
        if not line:
            if start_line_no == -1:
                pass
            else:
                yield block_tuple_t(start_line_no, end_line_no)
                start_line_no = end_line_no = -1
        else:
            if start_line_no == -1:
                start_line_no = cur_line_no
                end_line_no = cur_line_no + 1
            else:
                end_line_no += 1
    if start_line_no != -1:
        yield block_tuple_t(start_line_no, end_line_no)


def starts_with_blank(line: str):
    return line.startswith(' ') or line.startswith('\t')


def is_milestone(word: str) -> bool:
    return re.match(r'^[^=<>,;:~\[\]#(){}]+$', word) is not None


def promise_core(word: str) -> str:
    return word[8:-1]


def is_promise(word: str) -> bool:
    return word.startswith('promise[') and word.endswith(']') \
           and is_milestone(promise_core(word))


def is_date(word: str) -> bool:
    try:
        return bool(Date.fromisoformat(word))
    except:
        return False


def is_unknown_time(word: str) -> bool:
    return re.match(r'^\?+$', word) is not None


def is_estimate_time(word: str) -> bool:
    if re.match(r'^\d+(\.5)?d$', word):
        return True
    if re.match(r'^\d+h$', word):
        return True
    if word == '0':
        return True
    return False



from datetime import datetime as Datetime
from datetime import date as Date
from datetime import timedelta as Timedelta
import re
from typing import Protocol, Dict


class IDatetime(Protocol):
    hours_of_day: int = 4  # 一天工作日包含多少工作小时
    predict_minutes: int = 0  # 预警时间
    holiday_exclude_days = ''  # 节假日排除的日期
    holiday_include_days = ''  # 节假日包含的日期

    @classmethod
    def load_options(cls, options: Dict):
        if 'day' in options:
            day_hours_opt = options['day']
            assert day_hours_opt.endswith('h')
            cls.hours_of_day = int(day_hours_opt.rstrip('h'))
        if 'predict' in options:
            predict_opt = options['predict']
            cls.predict_minutes = cls.minute_of_str(predict_opt)
        if 'exclude' in options:
            cls.holiday_exclude_days = options['exclude']
        if 'include' in options:
            cls.holiday_include_days = options['include']

    @classmethod
    def is_holiday(cls, date: Date):
        is_weekend = 5 <= date.weekday() <= 6
        date_str = str(date)
        return (is_weekend and date_str not in cls.holiday_exclude_days) or date_str in cls.holiday_include_days

    @classmethod
    def noon(cls, datetime: Datetime) -> Datetime:
        return datetime.replace(hour=12, minute=0, second=0, microsecond=0)

    @classmethod
    def day_begin(cls, datetime: Datetime) -> Datetime:
        return cls.noon(datetime) - Timedelta(minutes=cls.hours_of_day * 60 / 2)

    @classmethod
    def day_end(cls, datetime: Datetime) -> Datetime:
        return cls.noon(datetime) + Timedelta(minutes=cls.hours_of_day * 60 / 2)

    @classmethod
    def day_to_next(cls, datetime: Datetime) -> Datetime:
        if datetime < cls.day_begin(datetime):
            return cls.day_begin(datetime)  # 基本不会使用，无效过多处理
        if datetime <= cls.day_end(datetime):
            return datetime
        left_minute = (datetime - cls.day_end(datetime)).seconds // 60
        tomorrow_begin = cls.day_begin(datetime) + Timedelta(days=1)
        return cls.add(tomorrow_begin, minutes=left_minute)

    @classmethod
    def skip_holiday(cls, datetime: Datetime, *, days: int) -> Datetime:
        date = datetime.date()
        if cls.is_holiday(date):
            return cls.skip_holiday(datetime + Timedelta(days=1), days=days)
        elif days <= 0:
            return datetime
        else:
            return cls.skip_holiday(datetime + Timedelta(days=1), days=days - 1)

    @classmethod
    def add(cls, datetime: Datetime, *, minutes: int) -> Datetime:
        if datetime < cls.day_begin(datetime):
            datetime = cls.day_begin(datetime)
        elif datetime > cls.day_end(datetime):
            datetime = cls.day_end(datetime)
        day_num, left_minute = minutes // (cls.hours_of_day * 60), minutes % (cls.hours_of_day * 60)
        next_day = cls.skip_holiday(datetime, days=day_num)
        return cls.day_to_next(next_day + Timedelta(minutes=left_minute))

    @classmethod
    def noon_of_str(cls, date_str: str) -> Datetime:
        day = Date.fromisoformat(date_str)
        return Datetime(day.year, day.month, day.day, 12, 0, 0)

    @classmethod
    def minute_of_str(cls, time_str: str) -> int:
        if re.match(r'^\d+d$', time_str):
            return int(time_str[:-1]) * cls.hours_of_day * 60
        if re.match(r'^\d+\.5d$', time_str):
            return int(time_str[:-3]) * cls.hours_of_day * 60 + 30
        if re.match(r'^\d+h$', time_str):
            return int(time_str[:-1]) * 60
        return 0



from unittest import TestCase
from pmo.syntax_util import split_into_blocks


class SyntaxUtilTest(TestCase):
    def test_split_into_blocks(self):
        lines = ["", "aa", " bbb", " cccc", "", "", "ddddd", "", "ee"]
        blocks = []
        for start, end in split_into_blocks(lines):
            blocks.append(','.join(lines[start: end]))
        self.assertListEqual(blocks, ['aa, bbb, cccc', 'ddddd', 'ee'])



from datetime import datetime as Datetime
from unittest import TestCase
from pmo.datetime_util import IDatetime


class DatetimeUtilTest(TestCase):
    def test_day_to_next(self):
        datetime = Datetime(2020, 11, 1, 15, 30, 0)
        datetime = IDatetime.day_to_next(datetime)
        self.assertEqual(str(datetime), '2020-11-02 11:30:00')

    def test_add(self):
        datetime = Datetime(2020, 11, 2, 11, 30, 0)
        datetime = IDatetime.add(datetime, minutes=90)
        self.assertEqual(str(datetime), '2020-11-02 13:00:00')
        datetime = IDatetime.add(datetime, minutes=90)
        self.assertEqual(str(datetime), '2020-11-03 10:30:00')



from datetime import datetime as Datetime
from typing import List, Optional, Union, NamedTuple, Dict
import requests
from pmo.estimate_time import EstimateTime, EstimateTimeTopic
from pmo.syntax_util import split_into_blocks, starts_with_blank, is_milestone, promise_core, is_promise, is_date, is_unknown_time, is_estimate_time
from pmo.datetime_util import IDatetime


class PmoSyntaxError(Exception):
    reason: str
    line_no: int

    def __init__(self, reason: str, line_no: int=0):
        self.reason = reason
        self.line_no = line_no

    def __str__(self):
        return f'Line {self.line_no+1}: {self.reason}'


class PmoGrammerError(Exception):
    reason: str

    def __init__(self, reason: str):
        self.reason = reason

    def __str__(self):
        return f'Error: {self.reason}'


class MilestoneKey(NamedTuple):
    name: str
    promise: bool

    def __str__(self) -> str:
        return f'promise[{self.name}]' if self.promise else self.name


class Project:
    blocks: List['Block']
    milestone_index: Dict[MilestoneKey, 'Milestone']
    block_index: Dict[MilestoneKey, 'Block']

    def __init__(self, lines: List[str]) -> None:
        self.blocks = []
        self.milestone_index = {}
        self.block_index = {}
        for start_line_no, end_line_no in split_into_blocks(lines):
            try:
                block = Block(lines[start_line_no: end_line_no])
                self.blocks.append(block)
                self.block_index[block.dependence_line.milestone.key] = block
            except PmoSyntaxError as e:
                e.line_no += start_line_no
                print(e)
                exit(1)

    def report(self, bot_key) -> None:
        report_lines = []
        milestones = list(m for m in self.milestone_index.values() if m.done_date is None)
        milestones.sort(key = lambda m: m.rank_weight, reverse=True)
        deadline_error = False
        last_task = ''
        report_lines.append('  will_finish < milestone < due_date')
        report_lines.append('----------------------------------------')
        now_at = Datetime.now()
        for milestone in milestones:
            safe_sign: str
            if milestone.will_finish is None:
                safe_sign = '✗'
            elif milestone.will_finish < now_at:
                safe_sign = '✗'
            elif milestone.due_date and milestone.will_finish > milestone.due_date:
                safe_sign = '✗'
            elif not milestone.due_date and milestone.promise:
                safe_sign = '✗'
            else:
                safe_sign = '✓'
            if safe_sign == '✗':
                deadline_error = True
            if milestone.will_finish:
                will_finish_str = ' ' + str(milestone.will_finish.date())
            else:
                will_finish_str = '~' + str(milestone.rank_weight.date())
            due_date_str = str(milestone.due_date.date()) if milestone.due_date else ' ' * len('2000-01-01')
            report_lines.append(f'{safe_sign} {will_finish_str} < {milestone.key} < {due_date_str}')
            last_task = milestone.key
        total = len(self.milestone_index)
        undone_count = len(milestones)
        report_lines.append('----------------------------------------')
        report_text = '\n'.join(report_lines)
        if bot_key and deadline_error:
            bot_report_text = report_text + '\n'
            if last_task:
                bot_report_text += f'您目前的任务是[{last_task}]，'
            bot_report_text += '任务计划异常，请尽快解决 :('
            url = 'https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=' + bot_key
            data = {"msgtype": "text", "text": {"content": bot_report_text}}
            resp = requests.post(url, json=data)
            if resp.status_code != 200 or resp.json()['errcode'] != 0:
                print('企业微信机器人消息通知失败 - ' + resp.text)
        print(report_text)
        if deadline_error:
            print(f'✗  {total - undone_count}/{total} milestones done.')
            exit(1)
        else:
            print(f'✓  {total - undone_count}/{total} milestones done.')

    def dfs_milestones(self) -> None:
        """
        从第一个block的milestone开始，计算所有milestone的预估完成时间
        """
        if not self.blocks:
            raise PmoGrammerError('至少需要一个milestone')
        try:
            for block in self.blocks:
                milestone: Milestone = block.dependence_line.milestone
                self.dfs(milestone.key)
        except PmoGrammerError as e:
            print(e)
            exit(1)

    def dfs(self, milestone_key: 'MilestoneKey') -> Optional['Milestone']:
        """
        计算milestone的预估完成时间，并且更新到milestone_index
        """
        if milestone_key not in self.block_index:
            return None
        if milestone_key in self.milestone_index:
            return self.milestone_index[milestone_key]
        all_dependence_done = True
        all_dependence_finishable = True
        will_finish: Datetime = Datetime(1970, 1, 1, 0, 0, 0)
        rank_weight: Datetime = Datetime.now()
        block = self.block_index[milestone_key]
        for dependence_key in block.dependence_line.dependences:
            if isinstance(dependence_key, Datetime):
                will_finish = max(will_finish, dependence_key)
                rank_weight = max(rank_weight, dependence_key)
            else:
                dependence: Optional[Milestone] = self.dfs(dependence_key)
                if dependence is None:
                    raise PmoGrammerError(f'"{dependence_key}"的依赖项未定义!')
                if dependence.done_date is None and dependence.will_finish is None:
                    all_dependence_finishable = False
                if dependence.done_date is None:
                    all_dependence_done = False
                    if dependence.will_finish:
                        will_finish = max(will_finish, dependence.will_finish)
                    rank_weight = max(rank_weight, dependence.rank_weight)
                else:
                    will_finish = max(will_finish, dependence.done_date)
                    rank_weight = max(rank_weight, dependence.done_date)
        # 依赖项没有预估完成时间 -> will_finish=None
        # 预估时长为.estimate & all_dependence_done -> 语法异常
        # 预估时长为.unknown & promise未完成 -> will_finish=None
        # 预估时长为.unknown & promise已完成 -> 语法异常
        # 预估时长为.unknown & promise=None -> 语法异常
        # 其他 -> will_finish += 总预估时长
        total_minute = 0
        for estimate_line in block.estimate_lines:
            estimate_time: 'EstimateTime' = estimate_line.estimate_time
            estimate_time_topic = estimate_time.topic()
            if estimate_time_topic == EstimateTimeTopic.estimate:
                if all_dependence_done:
                    raise PmoGrammerError(f'{milestone_key}：所有依赖项已完成，预估时长必须确定，不能是区间!')
                best_minutes, _ = estimate_time.payload(estimate_time_topic)
                total_minute += best_minutes
            elif estimate_time_topic == EstimateTimeTopic.unknown and not milestone_key.promise:
                promise_milestone: Optional['Milestone'] = self.dfs(MilestoneKey(milestone_key.name, promise=True))
                if promise_milestone is None:
                    raise PmoGrammerError(f'{milestone_key}：预估时长unknown，并且没有promise!' )
                elif promise_milestone.done_date is not None:
                    raise PmoGrammerError(f'{milestone_key}：promise已完成，预估时长必须确定!' )
                else:
                    total_minute = 99999999
                    rank_weight = max(rank_weight, promise_milestone.rank_weight)
            elif estimate_time_topic == EstimateTimeTopic.unknown:
                raise PmoGrammerError(f'{milestone_key}：promise里程碑的预估时长必须确定!')
            elif estimate_time_topic == EstimateTimeTopic.duration:
                total_minute += estimate_time.payload(estimate_time_topic)
            else:
                raise NotImplementedError
        milestone: Milestone = block.dependence_line.milestone
        if not all_dependence_finishable or will_finish.year == 1970 or total_minute >= 99999999:
            milestone.will_finish = None
        else:
            milestone.will_finish = IDatetime.add(will_finish, minutes=total_minute)
        milestone.rank_weight = IDatetime.add(rank_weight, minutes=60 if total_minute >= 99999999 else total_minute + 1)
        # 更新到milestone_index
        self.milestone_index[milestone_key] = milestone
        return milestone


class Block:
    dependence_line: 'DependenceLine'
    estimate_lines: List['EstimateLine']

    def __init__(self, lines: List[str]):
        if starts_with_blank(lines[0]):
            raise PmoSyntaxError('依赖定义行不能有缩进')
        self.dependence_line = DependenceLine(lines[0].strip())
        self.estimate_lines = []
        for line in lines[1:]:
            if not starts_with_blank(line):
                raise PmoSyntaxError('预估时长行需要有缩进')
            self.estimate_lines.append(EstimateLine(line.strip()))


class DependenceLine:
    milestone: 'Milestone'
    dependences: List[Union['MilestoneKey', Datetime]]

    def __init__(self, line: str) -> None:
        milestone_part, dependence_part = line.split(':', 1)
        self.milestone = Milestone(milestone_part.strip())
        self.dependences = []
        for dependence_word in dependence_part.strip().split():
            if is_date(dependence_word):
                self.dependences.append(IDatetime.noon_of_str(dependence_word))
            elif is_milestone(dependence_word):
                self.dependences.append(MilestoneKey(name=dependence_word, promise=False))
            elif is_promise(dependence_word):
                self.dependences.append(MilestoneKey(name=promise_core(dependence_word), promise=True))
            else:
                raise PmoSyntaxError(f'依赖项"{dependence_word}"不合法')


class EstimateLine:
    work: str
    estimate_time: 'EstimateTime'

    def __init__(self, line: str) -> None:
        work_part, est_word = line.split('=', 1)
        self.work = work_part
        est_word = est_word.strip()
        if is_estimate_time(est_word):
            self.estimate_time = EstimateTime(EstimateTimeTopic.duration, IDatetime.minute_of_str(est_word))
        elif is_unknown_time(est_word):
            self.estimate_time = EstimateTime(EstimateTimeTopic.unknown)
        elif est_word.count('~') == 1:
            first_seg, second_seg = est_word.split('~', 1)
            if is_estimate_time(first_seg) and is_estimate_time(second_seg):
                self.estimate_time = EstimateTime(EstimateTimeTopic.estimate, (IDatetime.minute_of_str(first_seg), IDatetime.minute_of_str(second_seg)))
            else:
                raise PmoSyntaxError(f'预估时长"{est_word}"不合法')
        else:
            raise PmoSyntaxError(f'预估时长"{est_word}"不合法')


class Milestone:
    name: str
    promise: bool
    will_finish: Optional[Datetime]  # 预估完成时间(乐观)
    due_date: Optional[Datetime]
    done_date: Optional[Datetime]
    rank_weight: Datetime  # 把.unknown当成1h的预估完成时间

    def __init__(self, word: str) -> None:
        self.will_finish = None
        self.due_date = None
        self.done_date = None
        if '=' in word:
            name_part, date_part = word.split('=', 1)
            date_part = date_part.strip()
            if not is_date(date_part):
                raise PmoSyntaxError(f'完成日期"{date_part}"不合法')
            self.done_date = IDatetime.noon_of_str(date_part)
        elif '<' in word:
            name_part, date_part = word.split('<', 1)
            date_part = date_part.strip()
            if not is_date(date_part):
                raise PmoSyntaxError(f'Deadline日期"{date_part}"不合法')
            self.due_date = IDatetime.noon_of_str(date_part.strip())
        else:
            name_part = word
        name_part = name_part.strip()
        if is_promise(name_part):
            self.name = promise_core(name_part)
            self.promise = True
        elif is_milestone(name_part):
            self.name = name_part
            self.promise = False
        else:
            raise PmoSyntaxError(f'依赖项"{name_part}"不合法')


    @property
    def key(self) -> MilestoneKey:
        return MilestoneKey(name=self.name, promise=self.promise)



import sys
from pmo.syntax import Project
from pmo.syntax_util import clear_comment
from pmo.datetime_util import IDatetime


def print_help():
    help_text = """Usage: pmomake [options] [target]
Options:
  --predict[=?d|?h]        预警时间，默认为0d
  --day[=?h]               一天工作日包含多少工作小时，默认为4h
  --exclude[=DATE]         排除的日期，DATE支持2020-10-01,2020-10-02这样的日期
  --include[=DATE]         包含的日期
  --bot[=string]           企业微信机器人key
"""
    print(help_text)


def parse_options():
    if len(sys.argv) <= 1 or '--help' in sys.argv or '-h' in sys.argv:
        print_help()
        exit(0)
    option_segs = sys.argv[1:-1]
    options = {}
    for seg in option_segs:
        if '=' not in seg or not seg.startswith('--'):
            print('错误的参数：' + seg)
            exit(1)
        alice, bob = seg.split('=', 1)
        options[alice.lstrip('-')] = bob
    return options


def entrypoint():
    options = parse_options()
    IDatetime.load_options(options)
    target = sys.argv[-1]
    lines = open(target).read().splitlines()
    lines_without_comment = [x for x in clear_comment(lines)]
    project = Project(lines_without_comment)
    project.dfs_milestones()
    bot_key = options.get('bot')
    project.report(bot_key)
